# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module contains a helper function to fill erfa.astrom struct and a
ScienceState, which allows to speed up coordinate transformations at the
expense of accuracy.
"""
import warnings

import erfa
import numpy as np

import astropy.units as u
from astropy.time import Time
from astropy.utils.exceptions import AstropyWarning
from astropy.utils.state import ScienceState

from .builtin_frames.utils import (
    get_cip,
    get_jd12,
    get_polar_motion,
    pav2pv,
    prepare_earth_position_vel,
)
from .matrix_utilities import rotation_matrix

__all__ = []


class ErfaAstrom:
    """
    The default provider for astrometry values.
    A utility class to extract the necessary arguments for
    erfa functions from frame attributes, call the corresponding
    erfa functions and return the astrom object.
    """

    @staticmethod
    def apco(frame_or_coord):
        """
        Wrapper for ``erfa.apco``, used in conversions AltAz <-> ICRS and CIRS <-> ICRS.

        Parameters
        ----------
        frame_or_coord : ``astropy.coordinates.BaseCoordinateFrame`` or ``astropy.coordinates.SkyCoord``
            Frame or coordinate instance in the corresponding frame
            for which to calculate the calculate the astrom values.
            For this function, an AltAz or CIRS frame is expected.
        """
        lon, lat, height = frame_or_coord.location.to_geodetic("WGS84")
        obstime = frame_or_coord.obstime

        jd1_tt, jd2_tt = get_jd12(obstime, "tt")
        xp, yp = get_polar_motion(obstime)
        sp = erfa.sp00(jd1_tt, jd2_tt)
        x, y, s = get_cip(jd1_tt, jd2_tt)
        era = erfa.era00(*get_jd12(obstime, "ut1"))
        earth_pv, earth_heliocentric = prepare_earth_position_vel(obstime)

        # refraction constants
        if hasattr(frame_or_coord, "pressure"):
            # this is an AltAz like frame. Calculate refraction
            refa, refb = erfa.refco(
                frame_or_coord.pressure.to_value(u.hPa),
                frame_or_coord.temperature.to_value(u.deg_C),
                frame_or_coord.relative_humidity.value,
                frame_or_coord.obswl.to_value(u.micron),
            )
        else:
            # This is not an AltAz frame, so don't bother computing refraction
            refa, refb = 0.0, 0.0

        return erfa.apco(
            jd1_tt,
            jd2_tt,
            earth_pv,
            earth_heliocentric,
            x,
            y,
            s,
            era,
            lon.to_value(u.radian),
            lat.to_value(u.radian),
            height.to_value(u.m),
            xp,
            yp,
            sp,
            refa,
            refb,
        )

    @staticmethod
    def apcs(frame_or_coord):
        """
        Wrapper for ``erfa.apcs``, used in conversions GCRS <-> ICRS.

        Parameters
        ----------
        frame_or_coord : ``astropy.coordinates.BaseCoordinateFrame`` or ``astropy.coordinates.SkyCoord``
            Frame or coordinate instance in the corresponding frame
            for which to calculate the calculate the astrom values.
            For this function, a GCRS frame is expected.
        """
        jd1_tt, jd2_tt = get_jd12(frame_or_coord.obstime, "tt")
        obs_pv = pav2pv(
            frame_or_coord.obsgeoloc.get_xyz(xyz_axis=-1).value,
            frame_or_coord.obsgeovel.get_xyz(xyz_axis=-1).value,
        )
        earth_pv, earth_heliocentric = prepare_earth_position_vel(
            frame_or_coord.obstime
        )
        return erfa.apcs(jd1_tt, jd2_tt, obs_pv, earth_pv, earth_heliocentric)

    @staticmethod
    def apio(frame_or_coord):
        """
        Slightly modified equivalent of ``erfa.apio``, used in conversions AltAz <-> CIRS.

        Since we use a topocentric CIRS frame, we have dropped the steps needed to calculate
        diurnal aberration.

        Parameters
        ----------
        frame_or_coord : ``astropy.coordinates.BaseCoordinateFrame`` or ``astropy.coordinates.SkyCoord``
            Frame or coordinate instance in the corresponding frame
            for which to calculate the calculate the astrom values.
            For this function, an AltAz frame is expected.
        """
        # Calculate erfa.apio input parameters.
        # TIO locator s'
        sp = erfa.sp00(*get_jd12(frame_or_coord.obstime, "tt"))

        # Earth rotation angle.
        theta = erfa.era00(*get_jd12(frame_or_coord.obstime, "ut1"))

        # Longitude and latitude in radians.
        lon, lat, height = frame_or_coord.location.to_geodetic("WGS84")
        elong = lon.to_value(u.radian)
        phi = lat.to_value(u.radian)

        # Polar motion, rotated onto local meridian
        xp, yp = get_polar_motion(frame_or_coord.obstime)

        # we need an empty astrom structure before we fill in the required sections
        astrom = np.zeros(frame_or_coord.obstime.shape, dtype=erfa.dt_eraASTROM)

        # Form the rotation matrix, CIRS to apparent [HA,Dec].
        r = (
            rotation_matrix(elong, "z", unit=u.radian)
            @ rotation_matrix(-yp, "x", unit=u.radian)
            @ rotation_matrix(-xp, "y", unit=u.radian)
            @ rotation_matrix(theta + sp, "z", unit=u.radian)
        )

        # Solve for local Earth rotation angle.
        a = r[..., 0, 0]
        b = r[..., 0, 1]
        eral = np.arctan2(b, a)
        astrom["eral"] = eral

        # Solve for polar motion [X,Y] with respect to local meridian.
        c = r[..., 0, 2]
        astrom["xpl"] = np.arctan2(c, np.sqrt(a * a + b * b))
        a = r[..., 1, 2]
        b = r[..., 2, 2]
        astrom["ypl"] = -np.arctan2(a, b)

        # Adjusted longitude.
        astrom["along"] = erfa.anpm(eral - theta)

        # Functions of latitude.
        astrom["sphi"] = np.sin(phi)
        astrom["cphi"] = np.cos(phi)

        # Omit two steps that are zero for a geocentric observer:
        # Observer's geocentric position and velocity (m, m/s, CIRS).
        # Magnitude of diurnal aberration vector.

        # Refraction constants.
        astrom["refa"], astrom["refb"] = erfa.refco(
            frame_or_coord.pressure.to_value(u.hPa),
            frame_or_coord.temperature.to_value(u.deg_C),
            frame_or_coord.relative_humidity.value,
            frame_or_coord.obswl.to_value(u.micron),
        )
        return astrom


class ErfaAstromInterpolator(ErfaAstrom):
    """
    A provider for astrometry values that does not call erfa
    for each individual timestamp but interpolates linearly
    between support points.

    For the interpolation, float64 MJD values are used, so time precision
    for the interpolation will be around a microsecond.

    This can dramatically speed up coordinate transformations,
    e.g. between CIRS and ICRS,
    when obstime is an array of many values (factors of 10 to > 100 depending
    on the selected resolution, number of points and the time range of the values).

    The precision of the transformation will still be in the order of microseconds
    for reasonable values of time_resolution, e.g. ``300 * u.s``.

    Users should benchmark performance and accuracy with the default transformation
    for their specific use case and then choose a suitable ``time_resolution``
    from there.

    This class is intended be used together with the ``erfa_astrom`` science state,
    e.g. in a context manager like this

    Example
    -------
    >>> from astropy.coordinates import SkyCoord, CIRS
    >>> from astropy.coordinates.erfa_astrom import erfa_astrom, ErfaAstromInterpolator
    >>> import astropy.units as u
    >>> from astropy.time import Time
    >>> import numpy as np

    >>> obstime = Time('2010-01-01T20:00:00') + np.linspace(0, 4, 1000) * u.hour
    >>> crab = SkyCoord(ra='05h34m31.94s', dec='22d00m52.2s')
    >>> with erfa_astrom.set(ErfaAstromInterpolator(300 * u.s)):
    ...    cirs = crab.transform_to(CIRS(obstime=obstime))
    """

    @u.quantity_input(time_resolution=u.day)
    def __init__(self, time_resolution):
        if time_resolution.to_value(u.us) < 10:
            warnings.warn(
                f"Using {self.__class__.__name__} with `time_resolution`"
                " below 10 microseconds might lead to numerical inaccuracies"
                " as the MJD-based interpolation is limited by floating point "
                " precision to about a microsecond of precision",
                AstropyWarning,
            )
        self.mjd_resolution = time_resolution.to_value(u.day)

    def _get_support_points(self, obstime):
        """
        Calculate support points for the interpolation.

        We divide the MJD by the time resolution (as single float64 values),
        and calculate ceil and floor.
        Then we take the unique and sorted values and scale back to MJD.
        This will create a sparse support for non-regular input obstimes.
        """
        mjd_scaled = np.ravel(obstime.mjd / self.mjd_resolution)

        # unique already does sorting
        mjd_u = np.unique(np.concatenate([np.floor(mjd_scaled), np.ceil(mjd_scaled)]))

        return Time(
            mjd_u * self.mjd_resolution,
            format="mjd",
            scale=obstime.scale,
        )

    @staticmethod
    def _prepare_earth_position_vel(support, obstime):
        """
        Calculate Earth's position and velocity.

        Uses the coarser grid ``support`` to do the calculation, and interpolates
        onto the finer grid ``obstime``.
        """
        pv_support, heliocentric_support = prepare_earth_position_vel(support)

        # do interpolation
        earth_pv = np.empty(obstime.shape, dtype=erfa.dt_pv)
        earth_heliocentric = np.empty(obstime.shape + (3,))
        for dim in range(3):
            for key in "pv":
                earth_pv[key][..., dim] = np.interp(
                    obstime.mjd, support.mjd, pv_support[key][..., dim]
                )
            earth_heliocentric[..., dim] = np.interp(
                obstime.mjd, support.mjd, heliocentric_support[..., dim]
            )

        return earth_pv, earth_heliocentric

    @staticmethod
    def _get_c2i(support, obstime):
        """
        Calculate the Celestial-to-Intermediate rotation matrix.

        Uses the coarser grid ``support`` to do the calculation, and interpolates
        onto the finer grid ``obstime``.
        """
        jd1_tt_support, jd2_tt_support = get_jd12(support, "tt")
        c2i_support = erfa.c2i06a(jd1_tt_support, jd2_tt_support)
        c2i = np.empty(obstime.shape + (3, 3))
        for dim1 in range(3):
            for dim2 in range(3):
                c2i[..., dim1, dim2] = np.interp(
                    obstime.mjd, support.mjd, c2i_support[..., dim1, dim2]
                )
        return c2i

    @staticmethod
    def _get_cip(support, obstime):
        """
        Find the X, Y coordinates of the CIP and the CIO locator, s.

        Uses the coarser grid ``support`` to do the calculation, and interpolates
        onto the finer grid ``obstime``.
        """
        jd1_tt_support, jd2_tt_support = get_jd12(support, "tt")
        cip_support = get_cip(jd1_tt_support, jd2_tt_support)
        return tuple(
            np.interp(obstime.mjd, support.mjd, cip_component)
            for cip_component in cip_support
        )

    @staticmethod
    def _get_polar_motion(support, obstime):
        """
        Find the two polar motion components in radians.

        Uses the coarser grid ``support`` to do the calculation, and interpolates
        onto the finer grid ``obstime``.
        """
        polar_motion_support = get_polar_motion(support)
        return tuple(
            np.interp(obstime.mjd, support.mjd, polar_motion_component)
            for polar_motion_component in polar_motion_support
        )

    def apco(self, frame_or_coord):
        """
        Wrapper for ``erfa.apco``, used in conversions AltAz <-> ICRS and CIRS <-> ICRS.

        Parameters
        ----------
        frame_or_coord : ``astropy.coordinates.BaseCoordinateFrame`` or ``astropy.coordinates.SkyCoord``
            Frame or coordinate instance in the corresponding frame
            for which to calculate the calculate the astrom values.
            For this function, an AltAz or CIRS frame is expected.
        """
        lon, lat, height = frame_or_coord.location.to_geodetic("WGS84")
        obstime = frame_or_coord.obstime
        support = self._get_support_points(obstime)
        jd1_tt, jd2_tt = get_jd12(obstime, "tt")

        # get the position and velocity arrays for the observatory.  Need to
        # have xyz in last dimension, and pos/vel in one-but-last.
        earth_pv, earth_heliocentric = self._prepare_earth_position_vel(
            support, obstime
        )

        xp, yp = self._get_polar_motion(support, obstime)
        sp = erfa.sp00(jd1_tt, jd2_tt)
        x, y, s = self._get_cip(support, obstime)
        era = erfa.era00(*get_jd12(obstime, "ut1"))

        # refraction constants
        if hasattr(frame_or_coord, "pressure"):
            # an AltAz like frame. Include refraction
            refa, refb = erfa.refco(
                frame_or_coord.pressure.to_value(u.hPa),
                frame_or_coord.temperature.to_value(u.deg_C),
                frame_or_coord.relative_humidity.value,
                frame_or_coord.obswl.to_value(u.micron),
            )
        else:
            # a CIRS like frame - no refraction
            refa, refb = 0.0, 0.0

        return erfa.apco(
            jd1_tt,
            jd2_tt,
            earth_pv,
            earth_heliocentric,
            x,
            y,
            s,
            era,
            lon.to_value(u.radian),
            lat.to_value(u.radian),
            height.to_value(u.m),
            xp,
            yp,
            sp,
            refa,
            refb,
        )

    def apcs(self, frame_or_coord):
        """
        Wrapper for ``erfa.apci``, used in conversions GCRS <-> ICRS.

        Parameters
        ----------
        frame_or_coord : ``astropy.coordinates.BaseCoordinateFrame`` or ``astropy.coordinates.SkyCoord``
            Frame or coordinate instance in the corresponding frame
            for which to calculate the calculate the astrom values.
            For this function, a GCRS frame is expected.
        """
        obstime = frame_or_coord.obstime
        support = self._get_support_points(obstime)

        # get the position and velocity arrays for the observatory.  Need to
        # have xyz in last dimension, and pos/vel in one-but-last.
        earth_pv, earth_heliocentric = self._prepare_earth_position_vel(
            support, obstime
        )
        pv = pav2pv(
            frame_or_coord.obsgeoloc.get_xyz(xyz_axis=-1).value,
            frame_or_coord.obsgeovel.get_xyz(xyz_axis=-1).value,
        )

        jd1_tt, jd2_tt = get_jd12(obstime, "tt")
        return erfa.apcs(jd1_tt, jd2_tt, pv, earth_pv, earth_heliocentric)


class erfa_astrom(ScienceState):
    """
    ScienceState to select with astrom provider is used in
    coordinate transformations.
    """

    _value = ErfaAstrom()

    @classmethod
    def validate(cls, value):
        if not isinstance(value, ErfaAstrom):
            raise TypeError(f"Must be an instance of {ErfaAstrom!r}")
        return value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from numpy.testing import assert_array_equal

from astropy.io.misc.asdf.tags.transform.basic import TransformType
from astropy.modeling import functional_models

from . import _parameter_to_value

__all__ = [
    "AiryDisk2DType",
    "Box1DType",
    "Box2DType",
    "Disk2DType",
    "Ellipse2DType",
    "Exponential1DType",
    "Gaussian1DType",
    "Gaussian2DType",
    "KingProjectedAnalytic1DType",
    "Logarithmic1DType",
    "Lorentz1DType",
    "Moffat1DType",
    "Moffat2DType",
    "Planar2D",
    "RedshiftScaleFactorType",
    "RickerWavelet1DType",
    "RickerWavelet2DType",
    "Ring2DType",
    "Sersic1DType",
    "Sersic2DType",
    "Sine1DType",
    "Cosine1DType",
    "Tangent1DType",
    "ArcSine1DType",
    "ArcCosine1DType",
    "ArcTangent1DType",
    "Trapezoid1DType",
    "TrapezoidDisk2DType",
    "Voigt1DType",
]


class AiryDisk2DType(TransformType):
    name = "transform/airy_disk2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.AiryDisk2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.AiryDisk2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            radius=node["radius"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "radius": _parameter_to_value(model.radius),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.AiryDisk2D) and isinstance(
            b, functional_models.AiryDisk2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.radius, b.radius)


class Box1DType(TransformType):
    name = "transform/box1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Box1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Box1D(
            amplitude=node["amplitude"], x_0=node["x_0"], width=node["width"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "width": _parameter_to_value(model.width),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Box1D) and isinstance(
            b, functional_models.Box1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.width, b.width)


class Box2DType(TransformType):
    name = "transform/box2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Box2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Box2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            x_width=node["x_width"],
            y_0=node["y_0"],
            y_width=node["y_width"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "x_width": _parameter_to_value(model.x_width),
            "y_0": _parameter_to_value(model.y_0),
            "y_width": _parameter_to_value(model.y_width),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Box2D) and isinstance(
            b, functional_models.Box2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.x_width, b.x_width)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.y_width, b.y_width)


class Disk2DType(TransformType):
    name = "transform/disk2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Disk2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Disk2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            R_0=node["R_0"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "R_0": _parameter_to_value(model.R_0),
        }

        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Disk2D) and isinstance(
            b, functional_models.Disk2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.R_0, b.R_0)


class Ellipse2DType(TransformType):
    name = "transform/ellipse2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Ellipse2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Ellipse2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            a=node["a"],
            b=node["b"],
            theta=node["theta"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "a": _parameter_to_value(model.a),
            "b": _parameter_to_value(model.b),
            "theta": _parameter_to_value(model.theta),
        }

        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Ellipse2D) and isinstance(
            b, functional_models.Ellipse2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.a, b.a)
        assert_array_equal(a.b, b.b)
        assert_array_equal(a.theta, b.theta)


class Exponential1DType(TransformType):
    name = "transform/exponential1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Exponential1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Exponential1D(
            amplitude=node["amplitude"], tau=node["tau"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "tau": _parameter_to_value(model.tau),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Exponential1D) and isinstance(
            b, functional_models.Exponential1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.tau, b.tau)


class Gaussian1DType(TransformType):
    name = "transform/gaussian1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Gaussian1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Gaussian1D(
            amplitude=node["amplitude"], mean=node["mean"], stddev=node["stddev"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "mean": _parameter_to_value(model.mean),
            "stddev": _parameter_to_value(model.stddev),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Gaussian1D) and isinstance(
            b, functional_models.Gaussian1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.mean, b.mean)
        assert_array_equal(a.stddev, b.stddev)


class Gaussian2DType(TransformType):
    name = "transform/gaussian2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Gaussian2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Gaussian2D(
            amplitude=node["amplitude"],
            x_mean=node["x_mean"],
            y_mean=node["y_mean"],
            x_stddev=node["x_stddev"],
            y_stddev=node["y_stddev"],
            theta=node["theta"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_mean": _parameter_to_value(model.x_mean),
            "y_mean": _parameter_to_value(model.y_mean),
            "x_stddev": _parameter_to_value(model.x_stddev),
            "y_stddev": _parameter_to_value(model.y_stddev),
            "theta": _parameter_to_value(model.theta),
        }

        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Gaussian2D) and isinstance(
            b, functional_models.Gaussian2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_mean, b.x_mean)
        assert_array_equal(a.y_mean, b.y_mean)
        assert_array_equal(a.x_stddev, b.x_stddev)
        assert_array_equal(a.y_stddev, b.y_stddev)
        assert_array_equal(a.theta, b.theta)


class KingProjectedAnalytic1DType(TransformType):
    name = "transform/king_projected_analytic1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.KingProjectedAnalytic1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.KingProjectedAnalytic1D(
            amplitude=node["amplitude"], r_core=node["r_core"], r_tide=node["r_tide"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "r_core": _parameter_to_value(model.r_core),
            "r_tide": _parameter_to_value(model.r_tide),
        }

        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.KingProjectedAnalytic1D) and isinstance(
            b, functional_models.KingProjectedAnalytic1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.r_core, b.r_core)
        assert_array_equal(a.r_tide, b.r_tide)


class Logarithmic1DType(TransformType):
    name = "transform/logarithmic1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Logarithmic1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Logarithmic1D(
            amplitude=node["amplitude"], tau=node["tau"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "tau": _parameter_to_value(model.tau),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Logarithmic1D) and isinstance(
            b, functional_models.Logarithmic1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.tau, b.tau)


class Lorentz1DType(TransformType):
    name = "transform/lorentz1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Lorentz1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Lorentz1D(
            amplitude=node["amplitude"], x_0=node["x_0"], fwhm=node["fwhm"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "fwhm": _parameter_to_value(model.fwhm),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Lorentz1D) and isinstance(
            b, functional_models.Lorentz1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.fwhm, b.fwhm)


class Moffat1DType(TransformType):
    name = "transform/moffat1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Moffat1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Moffat1D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            gamma=node["gamma"],
            alpha=node["alpha"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "gamma": _parameter_to_value(model.gamma),
            "alpha": _parameter_to_value(model.alpha),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Moffat1D) and isinstance(
            b, functional_models.Moffat1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.gamma, b.gamma)
        assert_array_equal(a.alpha, b.alpha)


class Moffat2DType(TransformType):
    name = "transform/moffat2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Moffat2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Moffat2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            gamma=node["gamma"],
            alpha=node["alpha"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "gamma": _parameter_to_value(model.gamma),
            "alpha": _parameter_to_value(model.alpha),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Moffat2D) and isinstance(
            b, functional_models.Moffat2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.gamma, b.gamma)
        assert_array_equal(a.alpha, b.alpha)


class Planar2D(TransformType):
    name = "transform/planar2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Planar2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Planar2D(
            slope_x=node["slope_x"],
            slope_y=node["slope_y"],
            intercept=node["intercept"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "slope_x": _parameter_to_value(model.slope_x),
            "slope_y": _parameter_to_value(model.slope_y),
            "intercept": _parameter_to_value(model.intercept),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Planar2D) and isinstance(
            b, functional_models.Planar2D
        )
        assert_array_equal(a.slope_x, b.slope_x)
        assert_array_equal(a.slope_y, b.slope_y)
        assert_array_equal(a.intercept, b.intercept)


class RedshiftScaleFactorType(TransformType):
    name = "transform/redshift_scale_factor"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.RedshiftScaleFactor"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.RedshiftScaleFactor(z=node["z"])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {"z": _parameter_to_value(model.z)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.RedshiftScaleFactor) and isinstance(
            b, functional_models.RedshiftScaleFactor
        )
        assert_array_equal(a.z, b.z)


class RickerWavelet1DType(TransformType):
    name = "transform/ricker_wavelet1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.RickerWavelet1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.RickerWavelet1D(
            amplitude=node["amplitude"], x_0=node["x_0"], sigma=node["sigma"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "sigma": _parameter_to_value(model.sigma),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.RickerWavelet1D) and isinstance(
            b, functional_models.RickerWavelet1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.sigma, b.sigma)


class RickerWavelet2DType(TransformType):
    name = "transform/ricker_wavelet2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.RickerWavelet2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.RickerWavelet2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            sigma=node["sigma"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "sigma": _parameter_to_value(model.sigma),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.RickerWavelet2D) and isinstance(
            b, functional_models.RickerWavelet2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.sigma, b.sigma)


class Ring2DType(TransformType):
    name = "transform/ring2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Ring2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Ring2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            r_in=node["r_in"],
            width=node["width"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "r_in": _parameter_to_value(model.r_in),
            "width": _parameter_to_value(model.width),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Ring2D) and isinstance(
            b, functional_models.Ring2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.r_in, b.r_in)
        assert_array_equal(a.width, b.width)


class Sersic1DType(TransformType):
    name = "transform/sersic1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Sersic1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Sersic1D(
            amplitude=node["amplitude"], r_eff=node["r_eff"], n=node["n"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "r_eff": _parameter_to_value(model.r_eff),
            "n": _parameter_to_value(model.n),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Sersic1D) and isinstance(
            b, functional_models.Sersic1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.r_eff, b.r_eff)
        assert_array_equal(a.n, b.n)


class Sersic2DType(TransformType):
    name = "transform/sersic2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Sersic2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Sersic2D(
            amplitude=node["amplitude"],
            r_eff=node["r_eff"],
            n=node["n"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            ellip=node["ellip"],
            theta=node["theta"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "r_eff": _parameter_to_value(model.r_eff),
            "n": _parameter_to_value(model.n),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "ellip": _parameter_to_value(model.ellip),
            "theta": _parameter_to_value(model.theta),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Sersic2D) and isinstance(
            b, functional_models.Sersic2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.r_eff, b.r_eff)
        assert_array_equal(a.n, b.n)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.ellip, b.ellip)
        assert_array_equal(a.theta, b.theta)


class Trigonometric1DType(TransformType):
    _model = None

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return cls._model(
            amplitude=node["amplitude"],
            frequency=node["frequency"],
            phase=node["phase"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "frequency": _parameter_to_value(model.frequency),
            "phase": _parameter_to_value(model.phase),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, cls._model) and isinstance(b, cls._model)
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.frequency, b.frequency)
        assert_array_equal(a.phase, b.phase)


class Sine1DType(Trigonometric1DType):
    name = "transform/sine1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Sine1D"]

    _model = functional_models.Sine1D


class Cosine1DType(Trigonometric1DType):
    name = "transform/cosine1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Cosine1D"]

    _model = functional_models.Cosine1D


class Tangent1DType(Trigonometric1DType):
    name = "transform/tangent1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Tangent1D"]

    _model = functional_models.Tangent1D


class ArcSine1DType(Trigonometric1DType):
    name = "transform/arcsine1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.ArcSine1D"]

    _model = functional_models.ArcSine1D


class ArcCosine1DType(Trigonometric1DType):
    name = "transform/arccosine1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.ArcCosine1D"]

    _model = functional_models.ArcCosine1D


class ArcTangent1DType(Trigonometric1DType):
    name = "transform/arctangent1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.ArcTangent1D"]

    _model = functional_models.ArcTangent1D


class Trapezoid1DType(TransformType):
    name = "transform/trapezoid1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Trapezoid1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Trapezoid1D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            width=node["width"],
            slope=node["slope"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "width": _parameter_to_value(model.width),
            "slope": _parameter_to_value(model.slope),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Trapezoid1D) and isinstance(
            b, functional_models.Trapezoid1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.width, b.width)
        assert_array_equal(a.slope, b.slope)


class TrapezoidDisk2DType(TransformType):
    name = "transform/trapezoid_disk2d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.TrapezoidDisk2D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.TrapezoidDisk2D(
            amplitude=node["amplitude"],
            x_0=node["x_0"],
            y_0=node["y_0"],
            R_0=node["R_0"],
            slope=node["slope"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "y_0": _parameter_to_value(model.y_0),
            "R_0": _parameter_to_value(model.R_0),
            "slope": _parameter_to_value(model.slope),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.TrapezoidDisk2D) and isinstance(
            b, functional_models.TrapezoidDisk2D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.y_0, b.y_0)
        assert_array_equal(a.R_0, b.R_0)
        assert_array_equal(a.slope, b.slope)


class Voigt1DType(TransformType):
    name = "transform/voigt1d"
    version = "1.0.0"
    types = ["astropy.modeling.functional_models.Voigt1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return functional_models.Voigt1D(
            x_0=node["x_0"],
            amplitude_L=node["amplitude_L"],
            fwhm_L=node["fwhm_L"],
            fwhm_G=node["fwhm_G"],
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "x_0": _parameter_to_value(model.x_0),
            "amplitude_L": _parameter_to_value(model.amplitude_L),
            "fwhm_L": _parameter_to_value(model.fwhm_L),
            "fwhm_G": _parameter_to_value(model.fwhm_G),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, functional_models.Voigt1D) and isinstance(
            b, functional_models.Voigt1D
        )
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.amplitude_L, b.amplitude_L)
        assert_array_equal(a.fwhm_L, b.fwhm_L)
        assert_array_equal(a.fwhm_G, b.fwhm_G)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Test sky projections defined in WCS Paper II"""
# pylint: disable=invalid-name, no-member
import os
import unittest.mock as mk

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

from astropy import units as u
from astropy import wcs
from astropy.io import fits
from astropy.modeling import projections
from astropy.modeling.parameters import InputParameterError
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename


def test_new_wcslib_projections():
    # Test that we are aware of all WCSLIB projections.
    # Detect if a new WCSLIB release introduced new projections.
    assert not set(wcs.PRJ_CODES).symmetric_difference(
        projections.projcodes + projections._NOT_SUPPORTED_PROJ_CODES
    )


def test_Projection_properties():
    projection = projections.Sky2Pix_PlateCarree()
    assert projection.n_inputs == 2
    assert projection.n_outputs == 2


PIX_COORDINATES = [-10, 30]

MAPS_DIR = os.path.join(os.pardir, os.pardir, "wcs", "tests", "data", "maps")

pars = [(x,) for x in projections.projcodes]
# There is no groundtruth file for the XPH projection available here:
#   https://www.atnf.csiro.au/people/mcalabre/WCS/example_data.html
pars.remove(("XPH",))


@pytest.mark.parametrize(("code",), pars)
def test_Sky2Pix(code):
    """Check astropy model eval against wcslib eval"""

    wcs_map = os.path.join(MAPS_DIR, f"1904-66_{code}.hdr")
    test_file = get_pkg_data_filename(wcs_map)
    header = fits.Header.fromfile(test_file, endcard=False, padding=False)

    params = []
    for i in range(3):
        key = f"PV2_{i + 1}"
        if key in header:
            params.append(header[key])

    w = wcs.WCS(header)
    w.wcs.crval = [0.0, 0.0]
    w.wcs.crpix = [0, 0]
    w.wcs.cdelt = [1, 1]
    wcslibout = w.wcs.p2s([PIX_COORDINATES], 1)
    wcs_pix = w.wcs.s2p(wcslibout["world"], 1)["pixcrd"]
    model = getattr(projections, "Sky2Pix_" + code)
    tinv = model(*params)
    x, y = tinv(wcslibout["phi"], wcslibout["theta"])
    assert_almost_equal(np.asarray(x), wcs_pix[:, 0])
    assert_almost_equal(np.asarray(y), wcs_pix[:, 1])

    assert isinstance(tinv.prjprm, wcs.Prjprm)


@pytest.mark.parametrize(("code",), pars)
def test_Pix2Sky(code):
    """Check astropy model eval against wcslib eval"""

    wcs_map = os.path.join(MAPS_DIR, f"1904-66_{code}.hdr")
    test_file = get_pkg_data_filename(wcs_map)
    header = fits.Header.fromfile(test_file, endcard=False, padding=False)

    params = []
    for i in range(3):
        key = f"PV2_{i + 1}"
        if key in header:
            params.append(header[key])

    w = wcs.WCS(header)
    w.wcs.crval = [0.0, 0.0]
    w.wcs.crpix = [0, 0]
    w.wcs.cdelt = [1, 1]
    wcslibout = w.wcs.p2s([PIX_COORDINATES], 1)
    wcs_phi = wcslibout["phi"]
    wcs_theta = wcslibout["theta"]
    model = getattr(projections, "Pix2Sky_" + code)
    tanprj = model(*params)
    phi, theta = tanprj(*PIX_COORDINATES)
    assert_almost_equal(np.asarray(phi), wcs_phi)
    assert_almost_equal(np.asarray(theta), wcs_theta)


@pytest.mark.parametrize(("code",), pars)
def test_Sky2Pix_unit(code):
    """Check astropy model eval against wcslib eval"""

    wcs_map = os.path.join(MAPS_DIR, f"1904-66_{code}.hdr")
    test_file = get_pkg_data_filename(wcs_map)
    header = fits.Header.fromfile(test_file, endcard=False, padding=False)

    params = []
    for i in range(3):
        key = f"PV2_{i + 1}"
        if key in header:
            params.append(header[key])

    w = wcs.WCS(header)
    w.wcs.crval = [0.0, 0.0]
    w.wcs.crpix = [0, 0]
    w.wcs.cdelt = [1, 1]
    wcslibout = w.wcs.p2s([PIX_COORDINATES], 1)
    wcs_pix = w.wcs.s2p(wcslibout["world"], 1)["pixcrd"]
    model = getattr(projections, "Sky2Pix_" + code)
    tinv = model(*params)
    x, y = tinv(wcslibout["phi"] * u.deg, wcslibout["theta"] * u.deg)
    assert_quantity_allclose(x, wcs_pix[:, 0] * u.deg)
    assert_quantity_allclose(y, wcs_pix[:, 1] * u.deg)


@pytest.mark.parametrize(("code",), pars)
def test_Pix2Sky_unit(code):
    """Check astropy model eval against wcslib eval"""

    wcs_map = os.path.join(MAPS_DIR, f"1904-66_{code}.hdr")
    test_file = get_pkg_data_filename(wcs_map)
    header = fits.Header.fromfile(test_file, endcard=False, padding=False)

    params = []
    for i in range(3):
        key = f"PV2_{i + 1}"
        if key in header:
            params.append(header[key])

    w = wcs.WCS(header)
    w.wcs.crval = [0.0, 0.0]
    w.wcs.crpix = [0, 0]
    w.wcs.cdelt = [1, 1]
    wcslibout = w.wcs.p2s([PIX_COORDINATES], 1)
    wcs_phi = wcslibout["phi"]
    wcs_theta = wcslibout["theta"]
    model = getattr(projections, "Pix2Sky_" + code)
    tanprj = model(*params)
    phi, theta = tanprj(*PIX_COORDINATES * u.deg)
    assert_quantity_allclose(phi, wcs_phi * u.deg)
    assert_quantity_allclose(theta, wcs_theta * u.deg)
    phi, theta = tanprj(*(PIX_COORDINATES * u.deg).to(u.rad))
    assert_quantity_allclose(phi, wcs_phi * u.deg)
    assert_quantity_allclose(theta, wcs_theta * u.deg)
    phi, theta = tanprj(*(PIX_COORDINATES * u.deg).to(u.arcmin))
    assert_quantity_allclose(phi, wcs_phi * u.deg)
    assert_quantity_allclose(theta, wcs_theta * u.deg)


@pytest.mark.parametrize(("code",), pars)
def test_projection_default(code):
    """Check astropy model eval with default parameters"""
    # Just makes sure that the default parameter values are reasonable
    # and accepted by wcslib.

    model = getattr(projections, "Sky2Pix_" + code)
    tinv = model()
    x, y = tinv(45, 45)

    model = getattr(projections, "Pix2Sky_" + code)
    tinv = model()
    x, y = tinv(0, 0)


class TestZenithalPerspective:
    """Test Zenithal Perspective projection"""

    def setup_class(self):
        ID = "AZP"
        wcs_map = os.path.join(MAPS_DIR, f"1904-66_{ID}.hdr")
        test_file = get_pkg_data_filename(wcs_map)
        header = fits.Header.fromfile(test_file, endcard=False, padding=False)
        self.wazp = wcs.WCS(header)
        self.wazp.wcs.crpix = np.array([0.0, 0.0])
        self.wazp.wcs.crval = np.array([0.0, 0.0])
        self.wazp.wcs.cdelt = np.array([1.0, 1.0])
        self.pv_kw = [kw[2] for kw in self.wazp.wcs.get_pv()]
        self.azp = projections.Pix2Sky_ZenithalPerspective(*self.pv_kw)

    def test_AZP_p2s(self):
        wcslibout = self.wazp.wcs.p2s([[-10, 30]], 1)
        wcs_phi = wcslibout["phi"]
        wcs_theta = wcslibout["theta"]
        phi, theta = self.azp(-10, 30)
        assert_almost_equal(np.asarray(phi), wcs_phi)
        assert_almost_equal(np.asarray(theta), wcs_theta)

    def test_AZP_s2p(self):
        wcslibout = self.wazp.wcs.p2s([[-10, 30]], 1)
        wcs_pix = self.wazp.wcs.s2p(wcslibout["world"], 1)["pixcrd"]
        x, y = self.azp.inverse(wcslibout["phi"], wcslibout["theta"])
        assert_almost_equal(np.asarray(x), wcs_pix[:, 0])
        assert_almost_equal(np.asarray(y), wcs_pix[:, 1])

    def test_validate(self):
        MESSAGE = r"Zenithal perspective projection is not defined for mu = -1"

        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Pix2Sky_ZenithalPerspective(-1)

        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Sky2Pix_ZenithalPerspective(-1)

        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Pix2Sky_SlantZenithalPerspective(-1)

        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Sky2Pix_SlantZenithalPerspective(-1)


class TestCylindricalPerspective:
    """Test cylindrical perspective projection"""

    def setup_class(self):
        ID = "CYP"
        wcs_map = os.path.join(MAPS_DIR, f"1904-66_{ID}.hdr")
        test_file = get_pkg_data_filename(wcs_map)
        header = fits.Header.fromfile(test_file, endcard=False, padding=False)
        self.wazp = wcs.WCS(header)
        self.wazp.wcs.crpix = np.array([0.0, 0.0])
        self.wazp.wcs.crval = np.array([0.0, 0.0])
        self.wazp.wcs.cdelt = np.array([1.0, 1.0])
        self.pv_kw = [kw[2] for kw in self.wazp.wcs.get_pv()]
        self.azp = projections.Pix2Sky_CylindricalPerspective(*self.pv_kw)

    def test_CYP_p2s(self):
        wcslibout = self.wazp.wcs.p2s([[-10, 30]], 1)
        wcs_phi = wcslibout["phi"]
        wcs_theta = wcslibout["theta"]
        phi, theta = self.azp(-10, 30)
        assert_almost_equal(np.asarray(phi), wcs_phi)
        assert_almost_equal(np.asarray(theta), wcs_theta)

    def test_CYP_s2p(self):
        wcslibout = self.wazp.wcs.p2s([[-10, 30]], 1)
        wcs_pix = self.wazp.wcs.s2p(wcslibout["world"], 1)["pixcrd"]
        x, y = self.azp.inverse(wcslibout["phi"], wcslibout["theta"])
        assert_almost_equal(np.asarray(x), wcs_pix[:, 0])
        assert_almost_equal(np.asarray(y), wcs_pix[:, 1])

    def test_validate(self):
        MESSAGE = r"CYP projection is not defined for .*"
        MESSAGE0 = r"CYP projection is not defined for mu = -lambda"
        MESSAGE1 = r"CYP projection is not defined for lambda = -mu"

        # Pix2Sky_CylindricalPerspective
        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Pix2Sky_CylindricalPerspective(1, -1)
        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Pix2Sky_CylindricalPerspective(-1, 1)

        model = projections.Pix2Sky_CylindricalPerspective()
        with pytest.raises(InputParameterError, match=MESSAGE0):
            model.mu = -1
        with pytest.raises(InputParameterError, match=MESSAGE1):
            model.lam = -1

        # Sky2Pix_CylindricalPerspective
        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Sky2Pix_CylindricalPerspective(1, -1)
        with pytest.raises(InputParameterError, match=MESSAGE):
            projections.Sky2Pix_CylindricalPerspective(-1, 1)

        model = projections.Sky2Pix_CylindricalPerspective()
        with pytest.raises(InputParameterError, match=MESSAGE0):
            model.mu = -1
        with pytest.raises(InputParameterError, match=MESSAGE1):
            model.lam = -1


def test_AffineTransformation2D():
    # Simple test with a scale and translation
    model = projections.AffineTransformation2D(
        matrix=[[2, 0], [0, 2]], translation=[1, 1]
    )

    # Coordinates for vertices of a rectangle
    rect = [[0, 0], [1, 0], [0, 3], [1, 3]]

    x, y = zip(*rect)

    new_rect = np.vstack(model(x, y)).T

    assert np.all(new_rect == [[1, 1], [3, 1], [1, 7], [3, 7]])

    # Matrix validation error
    MESSAGE = r"Expected transformation matrix to be a 2x2 array"
    with pytest.raises(InputParameterError, match=MESSAGE):
        model.matrix = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    # Translation validation error
    MESSAGE = (
        r"Expected translation vector to be a 2 element row or column vector array"
    )
    with pytest.raises(InputParameterError, match=MESSAGE):
        model.translation = [1, 2, 3]
    with pytest.raises(InputParameterError, match=MESSAGE):
        model.translation = [[1], [2]]
    with pytest.raises(InputParameterError, match=MESSAGE):
        model.translation = [[1, 2, 3]]

    # Incompatible shape error
    a = np.array([[1], [2], [3], [4]])
    b = a.ravel()
    with mk.patch.object(np, "vstack", autospec=True, side_effect=[a, b]) as mk_vstack:
        MESSAGE = r"Incompatible input shapes"
        with pytest.raises(ValueError, match=MESSAGE):
            model(x, y)
        with pytest.raises(ValueError, match=MESSAGE):
            model(x, y)
        assert mk_vstack.call_count == 2

    # Input shape evaluation error
    x = np.array([1, 2])
    y = np.array([1, 2, 3])
    MESSAGE = r"Expected input arrays to have the same shape"
    with pytest.raises(ValueError, match=MESSAGE):
        model.evaluate(x, y, model.matrix, model.translation)


def test_AffineTransformation2D_inverse():
    # Test non-invertible model
    model1 = projections.AffineTransformation2D(matrix=[[1, 1], [1, 1]])

    MESSAGE = r"Transformation matrix is singular; .* model does not have an inverse"
    with pytest.raises(InputParameterError, match=MESSAGE):
        model1.inverse

    model2 = projections.AffineTransformation2D(
        matrix=[[1.2, 3.4], [5.6, 7.8]], translation=[9.1, 10.11]
    )

    # Coordinates for vertices of a rectangle
    rect = [[0, 0], [1, 0], [0, 3], [1, 3]]

    x, y = zip(*rect)

    x_new, y_new = model2.inverse(*model2(x, y))
    assert_allclose([x, y], [x_new, y_new], atol=1e-10)

    model3 = projections.AffineTransformation2D(
        matrix=[[1.2, 3.4], [5.6, 7.8]] * u.m, translation=[9.1, 10.11] * u.m
    )

    x_new, y_new = model3.inverse(*model3(x * u.m, y * u.m))
    assert_allclose([x, y], [x_new, y_new], atol=1e-10)

    model4 = projections.AffineTransformation2D(
        matrix=[[1.2, 3.4], [5.6, 7.8]] * u.m, translation=[9.1, 10.11] * u.km
    )
    MESSAGE = r"matrix and translation must have the same units"
    with pytest.raises(ValueError, match=MESSAGE):
        model4.inverse(*model4(x * u.m, y * u.m))


def test_c_projection_striding():
    # This is just a simple test to make sure that the striding is
    # handled correctly in the projection C extension
    coords = np.arange(10).reshape((5, 2))

    model = projections.Sky2Pix_ZenithalPerspective(2, 30)

    phi, theta = model(coords[:, 0], coords[:, 1])

    assert_almost_equal(phi, [0.0, 2.2790416, 4.4889294, 6.6250643, 8.68301])

    assert_almost_equal(
        theta, [-76.4816918, -75.3594654, -74.1256332, -72.784558, -71.3406629]
    )


def test_c_projections_shaped():
    nx, ny = (5, 2)
    x = np.linspace(0, 1, nx)
    y = np.linspace(0, 1, ny)
    xv, yv = np.meshgrid(x, y)

    model = projections.Pix2Sky_TAN()

    phi, theta = model(xv, yv)

    assert_allclose(
        phi,
        [
            [0.0, 90.0, 90.0, 90.0, 90.0],
            [180.0, 165.96375653, 153.43494882, 143.13010235, 135.0],
        ],
    )

    assert_allclose(
        theta,
        [
            [90.0, 89.75000159, 89.50001269, 89.25004283, 89.00010152],
            [89.00010152, 88.96933478, 88.88210788, 88.75019826, 88.58607353],
        ],
    )


def test_affine_with_quantities():
    x = 1
    y = 2
    xdeg = (x * u.pix).to(u.deg, equivalencies=u.pixel_scale(2.5 * u.deg / u.pix))
    ydeg = (y * u.pix).to(u.deg, equivalencies=u.pixel_scale(2.5 * u.deg / u.pix))
    xpix = x * u.pix
    ypix = y * u.pix

    # test affine with matrix only
    qaff = projections.AffineTransformation2D(matrix=[[1, 2], [2, 1]] * u.deg)
    MESSAGE = (
        r"To use AffineTransformation with quantities, both matrix and unit need to be"
        r" quantities"
    )
    with pytest.raises(ValueError, match=MESSAGE):
        qx1, qy1 = qaff(
            xpix,
            ypix,
            equivalencies={
                "x": u.pixel_scale(2.5 * u.deg / u.pix),
                "y": u.pixel_scale(2.5 * u.deg / u.pix),
            },
        )

    # test affine with matrix and translation
    qaff = projections.AffineTransformation2D(
        matrix=[[1, 2], [2, 1]] * u.deg, translation=[1, 2] * u.deg
    )
    qx1, qy1 = qaff(
        xpix,
        ypix,
        equivalencies={
            "x": u.pixel_scale(2.5 * u.deg / u.pix),
            "y": u.pixel_scale(2.5 * u.deg / u.pix),
        },
    )
    aff = projections.AffineTransformation2D(
        matrix=[[1, 2], [2, 1]], translation=[1, 2]
    )
    x1, y1 = aff(xdeg.value, ydeg.value)
    assert_quantity_allclose(qx1, x1 * u.deg)
    assert_quantity_allclose(qy1, y1 * u.deg)

    # test the case of WCS PC and CDELT transformations
    pc = np.array(
        [
            [0.86585778922708, 0.50029020461607],
            [-0.50029020461607, 0.86585778922708],
        ]
    )
    cdelt = np.array(
        [
            [1, 3.0683055555556e-05],
            [3.0966944444444e-05, 1],
        ]
    )
    matrix = cdelt * pc
    qaff = projections.AffineTransformation2D(
        matrix=matrix * u.deg, translation=[0, 0] * u.deg
    )

    inv_matrix = np.linalg.inv(matrix)
    inv_qaff = projections.AffineTransformation2D(
        matrix=inv_matrix * u.pix, translation=[0, 0] * u.pix
    )
    qaff.inverse = inv_qaff
    qx1, qy1 = qaff(
        xpix,
        ypix,
        equivalencies={
            "x": u.pixel_scale(1 * u.deg / u.pix),
            "y": u.pixel_scale(1 * u.deg / u.pix),
        },
    )
    x1, y1 = qaff.inverse(
        qx1,
        qy1,
        equivalencies={
            "x": u.pixel_scale(1 * u.deg / u.pix),
            "y": u.pixel_scale(1 * u.deg / u.pix),
        },
    )
    assert_quantity_allclose(x1, xpix)
    assert_quantity_allclose(y1, ypix)


def test_Pix2Sky_ZenithalPerspective_inverse():
    model = projections.Pix2Sky_ZenithalPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ZenithalPerspective)
    assert inverse.mu == model.mu == 2
    assert_allclose(inverse.gamma, model.gamma)
    assert_allclose(inverse.gamma, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ZenithalPerspective_inverse():
    model = projections.Sky2Pix_ZenithalPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_AZP)
    assert inverse.mu == model.mu == 2
    assert_allclose(inverse.gamma, model.gamma)
    assert_allclose(inverse.gamma, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_SlantZenithalPerspective_inverse():
    model = projections.Pix2Sky_SlantZenithalPerspective(2, 30, 40)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_SlantZenithalPerspective)
    assert inverse.mu == model.mu == 2
    assert_allclose(inverse.phi0, model.phi0)
    assert_allclose(inverse.theta0, model.theta0)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_SlantZenithalPerspective_inverse():
    model = projections.Sky2Pix_SlantZenithalPerspective(2, 30, 40)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_SlantZenithalPerspective)
    assert inverse.mu == model.mu == 2
    assert_allclose(inverse.phi0, model.phi0)
    assert_allclose(inverse.theta0, model.theta0)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Gnomonic_inverse():
    model = projections.Pix2Sky_Gnomonic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Gnomonic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Gnomonic_inverse():
    model = projections.Sky2Pix_Gnomonic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Gnomonic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Stereographic_inverse():
    model = projections.Pix2Sky_Stereographic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Stereographic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Stereographic_inverse():
    model = projections.Sky2Pix_Stereographic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Stereographic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_SlantOrthographic_inverse():
    model = projections.Pix2Sky_SlantOrthographic(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_SlantOrthographic)
    assert inverse.xi == model.xi == 2
    assert inverse.eta == model.eta == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-8)
    assert_allclose(b, y, atol=1e-8)


def test_Sky2Pix_SlantOrthographic_inverse():
    model = projections.Sky2Pix_SlantOrthographic(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_SlantOrthographic)
    assert inverse.xi == model.xi == 2
    assert inverse.eta == model.eta == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-8)
    assert_allclose(b, y, atol=1e-8)


def test_Pix2Sky_ZenithalEquidistant_inverse():
    model = projections.Pix2Sky_ZenithalEquidistant()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ZenithalEquidistant)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ZenithalEquidistant_inverse():
    model = projections.Sky2Pix_ZenithalEquidistant()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ZenithalEquidistant)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_ZenithalEqualArea_inverse():
    model = projections.Pix2Sky_ZenithalEqualArea()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ZenithalEqualArea)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ZenithalEqualArea_inverse():
    model = projections.Sky2Pix_ZenithalEqualArea()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ZenithalEqualArea)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Airy_inverse():
    model = projections.Pix2Sky_Airy(30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Airy)
    assert inverse.theta_b == model.theta_b == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Airy_inverse():
    model = projections.Sky2Pix_Airy(30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Airy)
    assert inverse.theta_b == model.theta_b == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_CylindricalPerspective_inverse():
    model = projections.Pix2Sky_CylindricalPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_CylindricalPerspective)
    assert inverse.mu == model.mu == 2
    assert inverse.lam == model.lam == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_CylindricalPerspective_inverse():
    model = projections.Sky2Pix_CylindricalPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_CylindricalPerspective)
    assert inverse.mu == model.mu == 2
    assert inverse.lam == model.lam == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_CylindricalEqualArea_inverse():
    model = projections.Pix2Sky_CylindricalEqualArea(0.567)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_CylindricalEqualArea)
    assert inverse.lam == model.lam == 0.567


def test_Sky2Pix_CylindricalEqualArea_inverse():
    model = projections.Sky2Pix_CylindricalEqualArea(0.765)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_CylindricalEqualArea)
    assert inverse.lam == model.lam == 0.765


def test_Pix2Sky_PlateCarree_inverse():
    model = projections.Pix2Sky_PlateCarree()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_PlateCarree)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_PlateCarree_inverse():
    model = projections.Sky2Pix_PlateCarree()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_PlateCarree)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Mercator_inverse():
    model = projections.Pix2Sky_Mercator()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Mercator)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Mercator_inverse():
    model = projections.Sky2Pix_Mercator()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Mercator)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_SansonFlamsteed_inverse():
    model = projections.Pix2Sky_SansonFlamsteed()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_SansonFlamsteed)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_SansonFlamsteed_inverse():
    model = projections.Sky2Pix_SansonFlamsteed()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_SansonFlamsteed)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Parabolic_inverse():
    model = projections.Pix2Sky_Parabolic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Parabolic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Parabolic_inverse():
    model = projections.Sky2Pix_Parabolic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Parabolic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Molleweide_inverse():
    model = projections.Pix2Sky_Molleweide()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Molleweide)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Molleweide_inverse():
    model = projections.Sky2Pix_Molleweide()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Molleweide)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_HammerAitoff_inverse():
    model = projections.Pix2Sky_HammerAitoff()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_HammerAitoff)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_HammerAitoff_inverse():
    model = projections.Sky2Pix_HammerAitoff()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_HammerAitoff)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_ConicPerspective_inverse():
    model = projections.Pix2Sky_ConicPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ConicPerspective)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ConicPerspective_inverse():
    model = projections.Sky2Pix_ConicPerspective(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ConicPerspective)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_ConicEqualArea_inverse():
    model = projections.Pix2Sky_ConicEqualArea(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ConicEqualArea)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ConicEqualArea_inverse():
    model = projections.Sky2Pix_ConicEqualArea(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ConicEqualArea)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_ConicEquidistant_inverse():
    model = projections.Pix2Sky_ConicEquidistant(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ConicEquidistant)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ConicEquidistant_inverse():
    model = projections.Sky2Pix_ConicEquidistant(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ConicEquidistant)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_ConicOrthomorphic_inverse():
    model = projections.Pix2Sky_ConicOrthomorphic(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_ConicOrthomorphic)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_ConicOrthomorphic_inverse():
    model = projections.Sky2Pix_ConicOrthomorphic(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_ConicOrthomorphic)
    assert inverse.sigma == model.sigma == 2
    assert_allclose(inverse.delta, model.delta)
    assert_allclose(inverse.delta, 30)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_BonneEqualArea_inverse():
    model = projections.Pix2Sky_BonneEqualArea(2)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_BonneEqualArea)
    assert inverse.theta1 == model.theta1 == 2

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_BonneEqualArea_inverse():
    model = projections.Sky2Pix_BonneEqualArea(2)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_BonneEqualArea)
    assert inverse.theta1 == model.theta1 == 2

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_Polyconic_inverse():
    model = projections.Pix2Sky_Polyconic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_Polyconic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_Polyconic_inverse():
    model = projections.Sky2Pix_Polyconic()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_Polyconic)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_TangentialSphericalCube_inverse():
    model = projections.Pix2Sky_TangentialSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_TangentialSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_TangentialSphericalCube_inverse():
    model = projections.Sky2Pix_TangentialSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_TangentialSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_COBEQuadSphericalCube_inverse():
    model = projections.Pix2Sky_COBEQuadSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_COBEQuadSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-3)
    assert_allclose(b, y, atol=1e-3)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-3)
    assert_allclose(b, y, atol=1e-3)


def test_Sky2Pix_COBEQuadSphericalCube_inverse():
    model = projections.Sky2Pix_COBEQuadSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_COBEQuadSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-3)
    assert_allclose(b, y, atol=1e-3)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-3)
    assert_allclose(b, y, atol=1e-3)


def test_Pix2Sky_QuadSphericalCube_inverse():
    model = projections.Pix2Sky_QuadSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_QuadSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_QuadSphericalCube_inverse():
    model = projections.Sky2Pix_QuadSphericalCube()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_QuadSphericalCube)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_HEALPix_inverse():
    model = projections.Pix2Sky_HEALPix(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_HEALPix)
    assert inverse.H == model.H == 2
    assert inverse.X == model.X == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_HEALPix_inverse():
    model = projections.Sky2Pix_HEALPix(2, 30)
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_HEALPix)
    assert inverse.H == model.H == 2
    assert inverse.X == model.X == 30

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Pix2Sky_HEALPixPolar_inverse():
    model = projections.Pix2Sky_HEALPixPolar()
    inverse = model.inverse
    assert isinstance(inverse, projections.Sky2Pix_HEALPixPolar)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)


def test_Sky2Pix_HEALPixPolar_inverse():
    model = projections.Sky2Pix_HEALPixPolar()
    inverse = model.inverse
    assert isinstance(inverse, projections.Pix2Sky_HEALPixPolar)

    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 100)
    a, b = model(*inverse(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)
    a, b = inverse(*model(x, y))
    assert_allclose(a, x, atol=1e-12)
    assert_allclose(b, y, atol=1e-12)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# STDLIB
import json
import os

# THIRD PARTY
import pytest

# LOCAL
import astropy.units as u
from astropy.cosmology import units as cu
from astropy.cosmology.connect import readwrite_registry
from astropy.cosmology.core import Cosmology

from .base import ReadWriteDirectTestBase, ReadWriteTestMixinBase

###############################################################################


def read_json(filename, **kwargs):
    """Read JSON.

    Parameters
    ----------
    filename : str
    **kwargs
        Keyword arguments into :meth:`~astropy.cosmology.Cosmology.from_format`

    Returns
    -------
    `~astropy.cosmology.Cosmology` instance
    """
    # read
    if isinstance(filename, (str, bytes, os.PathLike)):
        with open(filename) as file:
            data = file.read()
    else:  # file-like : this also handles errors in dumping
        data = filename.read()

    mapping = json.loads(data)  # parse json mappable to dict

    # deserialize Quantity
    with u.add_enabled_units(cu.redshift):
        for k, v in mapping.items():
            if isinstance(v, dict) and "value" in v and "unit" in v:
                mapping[k] = u.Quantity(v["value"], v["unit"])
        for k, v in mapping.get("meta", {}).items():  # also the metadata
            if isinstance(v, dict) and "value" in v and "unit" in v:
                mapping["meta"][k] = u.Quantity(v["value"], v["unit"])

    return Cosmology.from_format(mapping, format="mapping", **kwargs)


def write_json(cosmology, file, *, overwrite=False):
    """Write Cosmology to JSON.

    Parameters
    ----------
    cosmology : `astropy.cosmology.Cosmology` subclass instance
    file : path-like or file-like
    overwrite : bool (optional, keyword-only)
    """
    data = cosmology.to_format("mapping")  # start by turning into dict
    data["cosmology"] = data["cosmology"].__qualname__

    # serialize Quantity
    for k, v in data.items():
        if isinstance(v, u.Quantity):
            data[k] = {"value": v.value.tolist(), "unit": str(v.unit)}
    for k, v in data.get("meta", {}).items():  # also serialize the metadata
        if isinstance(v, u.Quantity):
            data["meta"][k] = {"value": v.value.tolist(), "unit": str(v.unit)}

    # check that file exists and whether to overwrite.
    if os.path.exists(file) and not overwrite:
        raise OSError(f"{file} exists. Set 'overwrite' to write over.")
    with open(file, "w") as write_file:
        json.dump(data, write_file)


def json_identify(origin, filepath, fileobj, *args, **kwargs):
    return filepath is not None and filepath.endswith(".json")


###############################################################################


class ReadWriteJSONTestMixin(ReadWriteTestMixinBase):
    """
    Tests for a Cosmology[Read/Write] with ``format="json"``.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must dfine a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmology`` for an example.
    """

    @pytest.fixture(scope="class", autouse=True)
    def register_and_unregister_json(self):
        """Setup & teardown for JSON read/write tests."""
        # Register
        readwrite_registry.register_reader("json", Cosmology, read_json, force=True)
        readwrite_registry.register_writer("json", Cosmology, write_json, force=True)
        readwrite_registry.register_identifier(
            "json", Cosmology, json_identify, force=True
        )

        yield  # Run all tests in class

        # Unregister
        readwrite_registry.unregister_reader("json", Cosmology)
        readwrite_registry.unregister_writer("json", Cosmology)
        readwrite_registry.unregister_identifier("json", Cosmology)

    # ========================================================================

    def test_readwrite_json_subclass_partial_info(
        self, cosmo_cls, cosmo, read, write, tmp_path, add_cu
    ):
        """
        Test writing from an instance and reading from that class.
        This works with missing information.
        """
        fp = tmp_path / "test_readwrite_json_subclass_partial_info.json"

        # test write
        cosmo.write(fp, format="json")

        # partial information
        with open(fp) as file:
            L = file.readlines()[0]
        L = (
            L[: L.index('"cosmology":')] + L[L.index(", ") + 2 :]
        )  # remove cosmology  : #203
        i = L.index('"Tcmb0":')  # delete Tcmb0
        L = (
            L[:i] + L[L.index(", ", L.index(", ", i) + 1) + 2 :]
        )  # second occurrence  : #203

        tempfname = tmp_path / f"{cosmo.name}_temp.json"
        with open(tempfname, "w") as file:
            file.writelines([L])

        # read with the same class that wrote fills in the missing info with
        # the default value
        got = cosmo_cls.read(tempfname, format="json")
        got2 = read(tempfname, format="json", cosmology=cosmo_cls)
        got3 = read(tempfname, format="json", cosmology=cosmo_cls.__qualname__)

        assert (got == got2) and (got2 == got3)  # internal consistency

        # not equal, because Tcmb0 is changed, which also changes m_nu
        assert got != cosmo
        assert got.Tcmb0 == cosmo_cls._init_signature.parameters["Tcmb0"].default
        assert got.clone(name=cosmo.name, Tcmb0=cosmo.Tcmb0, m_nu=cosmo.m_nu) == cosmo
        # but the metadata is the same
        assert got.meta == cosmo.meta


class TestReadWriteJSON(ReadWriteDirectTestBase, ReadWriteJSONTestMixin):
    """
    Directly test ``read/write_json``.
    These are not public API and are discouraged from use, in favor of
    ``Cosmology.read/write(..., format="json")``, but should be
    tested regardless b/c they are used internally.
    """

    def setup_class(self):
        self.functions = {"read": read_json, "write": write_json}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from matplotlib import rc_context
from numpy.testing import assert_almost_equal

from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.units import UnitsError
from astropy.visualization.wcsaxes.formatter_locator import (
    AngleFormatterLocator,
    ScalarFormatterLocator,
)


class TestAngleFormatterLocator:
    def test_no_options(self):
        fl = AngleFormatterLocator()
        assert fl.values is None
        assert fl.number == 5
        assert fl.spacing is None

    def test_too_many_options(self):
        MESSAGE = r"At most one of values/number/spacing can be specified"

        with pytest.raises(ValueError, match=MESSAGE):
            AngleFormatterLocator(values=[1.0, 2.0], number=5)

        with pytest.raises(ValueError, match=MESSAGE):
            AngleFormatterLocator(values=[1.0, 2.0], spacing=5.0 * u.deg)

        with pytest.raises(ValueError, match=MESSAGE):
            AngleFormatterLocator(number=5, spacing=5.0 * u.deg)

        with pytest.raises(ValueError, match=MESSAGE):
            AngleFormatterLocator(values=[1.0, 2.0], number=5, spacing=5.0 * u.deg)

    def test_values(self):
        fl = AngleFormatterLocator(values=[0.1, 1.0, 14.0] * u.degree)
        assert fl.values.to_value(u.degree).tolist() == [0.1, 1.0, 14.0]
        assert fl.number is None
        assert fl.spacing is None

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(values.to_value(u.degree), [0.1, 1.0, 14.0])

    def test_number(self):
        fl = AngleFormatterLocator(number=7)
        assert fl.values is None
        assert fl.number == 7
        assert fl.spacing is None

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(values.to_value(u.degree), [35.0, 40.0, 45.0, 50.0, 55.0])

        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(
            values.to_value(u.degree), [34.5, 34.75, 35.0, 35.25, 35.5, 35.75, 36.0]
        )

        fl.format = "dd"
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.to_value(u.degree), [35.0, 36.0])

    def test_spacing(self):
        with pytest.raises(
            TypeError,
            match=(
                r"spacing should be an astropy.units.Quantity instance with units of"
                r" angle"
            ),
        ):
            AngleFormatterLocator(spacing=3.0)

        fl = AngleFormatterLocator(spacing=3.0 * u.degree)
        assert fl.values is None
        assert fl.number is None
        assert fl.spacing == 3.0 * u.degree

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(
            values.to_value(u.degree), [36.0, 39.0, 42.0, 45.0, 48.0, 51.0, 54.0]
        )

        fl.spacing = 30.0 * u.arcmin
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.to_value(u.degree), [34.5, 35.0, 35.5, 36.0])

        with pytest.warns(UserWarning, match=r"Spacing is too small"):
            fl.format = "dd"
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.to_value(u.degree), [35.0, 36.0])

    def test_minor_locator(self):
        fl = AngleFormatterLocator()

        values, spacing = fl.locator(34.3, 55.4)

        minor_values = fl.minor_locator(spacing, 5, 34.3, 55.4)

        assert_almost_equal(
            minor_values.to_value(u.degree),
            [
                36.0,
                37.0,
                38.0,
                39.0,
                41.0,
                42.0,
                43.0,
                44.0,
                46.0,
                47.0,
                48.0,
                49.0,
                51.0,
                52.0,
                53.0,
                54.0,
            ],
        )

        minor_values = fl.minor_locator(spacing, 2, 34.3, 55.4)

        assert_almost_equal(minor_values.to_value(u.degree), [37.5, 42.5, 47.5, 52.5])

        fl.values = [0.1, 1.0, 14.0] * u.degree

        values, spacing = fl.locator(34.3, 36.1)

        minor_values = fl.minor_locator(spacing, 2, 34.3, 55.4)

        assert_almost_equal(minor_values.to_value(u.degree), [])

    @pytest.mark.parametrize(
        ("format", "string"),
        [
            ("dd", "15\xb0"),
            ("dd:mm", "15\xb024'"),
            ("dd:mm:ss", "15\xb023'32\""),
            ("dd:mm:ss.s", "15\xb023'32.0\""),
            ("dd:mm:ss.ssss", "15\xb023'32.0316\""),
            ("hh", "1h"),
            ("hh:mm", "1h02m"),
            ("hh:mm:ss", "1h01m34s"),
            ("hh:mm:ss.s", "1h01m34.1s"),
            ("hh:mm:ss.ssss", "1h01m34.1354s"),
            ("d", "15\xb0"),
            ("d.d", "15.4\xb0"),
            ("d.dd", "15.39\xb0"),
            ("d.ddd", "15.392\xb0"),
            ("m", "924'"),
            ("m.m", "923.5'"),
            ("m.mm", "923.53'"),
            ("s", '55412"'),
            ("s.s", '55412.0"'),
            ("s.ss", '55412.03"'),
        ],
    )
    def test_format(self, format, string):
        fl = AngleFormatterLocator(number=5, format=format)
        assert fl.formatter([15.392231] * u.degree, None, format="ascii")[0] == string

    @pytest.mark.parametrize(
        ("separator", "format", "string"),
        [
            (("deg", "'", '"'), "dd", "15deg"),
            (("deg", "'", '"'), "dd:mm", "15deg24'"),
            (("deg", "'", '"'), "dd:mm:ss", "15deg23'32\""),
            ((":", "-", "s"), "dd:mm:ss.s", "15:23-32.0s"),
            (":", "dd:mm:ss.s", "15:23:32.0"),
            ((":", ":", "s"), "hh", "1:"),
            (("-", "-", "s"), "hh:mm:ss.ssss", "1-01-34.1354s"),
            (("d", ":", '"'), "d", "15\xb0"),
            (("d", ":", '"'), "d.d", "15.4\xb0"),
        ],
    )
    def test_separator(self, separator, format, string):
        fl = AngleFormatterLocator(number=5, format=format)
        fl.sep = separator
        assert fl.formatter([15.392231] * u.degree, None)[0] == string

    def test_latex_format(self):
        fl = AngleFormatterLocator(number=5, format="dd:mm:ss")
        assert fl.formatter([15.392231] * u.degree, None)[0] == "15\xb023'32\""
        with rc_context(rc={"text.usetex": True}):
            assert (
                fl.formatter([15.392231] * u.degree, None)[0]
                == "$15^\\circ23{}^\\prime32{}^{\\prime\\prime}$"
            )

    @pytest.mark.parametrize("format", ["x.xxx", "dd.ss", "dd:ss", "mdd:mm:ss"])
    def test_invalid_formats(self, format):
        fl = AngleFormatterLocator(number=5)
        with pytest.raises(ValueError, match=f"Invalid format: {format}"):
            fl.format = format

    @pytest.mark.parametrize(
        ("format", "base_spacing"),
        [
            ("dd", 1.0 * u.deg),
            ("dd:mm", 1.0 * u.arcmin),
            ("dd:mm:ss", 1.0 * u.arcsec),
            ("dd:mm:ss.ss", 0.01 * u.arcsec),
            ("hh", 15.0 * u.deg),
            ("hh:mm", 15.0 * u.arcmin),
            ("hh:mm:ss", 15.0 * u.arcsec),
            ("hh:mm:ss.ss", 0.15 * u.arcsec),
            ("d", 1.0 * u.deg),
            ("d.d", 0.1 * u.deg),
            ("d.dd", 0.01 * u.deg),
            ("d.ddd", 0.001 * u.deg),
            ("m", 1.0 * u.arcmin),
            ("m.m", 0.1 * u.arcmin),
            ("m.mm", 0.01 * u.arcmin),
            ("s", 1.0 * u.arcsec),
            ("s.s", 0.1 * u.arcsec),
            ("s.ss", 0.01 * u.arcsec),
        ],
    )
    def test_base_spacing(self, format, base_spacing):
        fl = AngleFormatterLocator(number=5, format=format)
        assert fl.base_spacing == base_spacing

    def test_incorrect_spacing(self):
        fl = AngleFormatterLocator()
        fl.spacing = 0.032 * u.deg
        with pytest.warns(
            UserWarning, match=r"Spacing is not a multiple of base spacing"
        ):
            fl.format = "dd:mm:ss"
        assert_almost_equal(fl.spacing.to_value(u.arcsec), 115.0)

    def test_decimal_values(self):
        # Regression test for a bug that meant that the spacing was not
        # determined correctly for decimal coordinates

        fl = AngleFormatterLocator()
        fl.format = "d.dddd"
        assert_quantity_allclose(
            fl.locator(266.9730, 266.9750)[0],
            [266.9735, 266.9740, 266.9745, 266.9750] * u.deg,
        )

        fl = AngleFormatterLocator(decimal=True, format_unit=u.hourangle, number=4)
        assert_quantity_allclose(
            fl.locator(266.9730, 266.9750)[0], [17.79825, 17.79830] * u.hourangle
        )

    def test_values_unit(self):
        # Make sure that the intrinsic unit and format unit are correctly
        # taken into account when using the locator

        fl = AngleFormatterLocator(unit=u.arcsec, format_unit=u.arcsec, decimal=True)
        assert_quantity_allclose(
            fl.locator(850, 2150)[0],
            [1000.0, 1200.0, 1400.0, 1600.0, 1800.0, 2000.0] * u.arcsec,
        )

        fl = AngleFormatterLocator(unit=u.arcsec, format_unit=u.degree, decimal=False)
        assert_quantity_allclose(
            fl.locator(850, 2150)[0], [15.0, 20.0, 25.0, 30.0, 35.0] * u.arcmin
        )

        fl = AngleFormatterLocator(
            unit=u.arcsec, format_unit=u.hourangle, decimal=False
        )
        assert_quantity_allclose(
            fl.locator(850, 2150)[0],
            [60.0, 75.0, 90.0, 105.0, 120.0, 135.0] * (15 * u.arcsec),
        )

        fl = AngleFormatterLocator(unit=u.arcsec)
        fl.format = "dd:mm:ss"
        assert_quantity_allclose(fl.locator(0.9, 1.1)[0], [1] * u.arcsec)

        fl = AngleFormatterLocator(unit=u.arcsec, spacing=0.2 * u.arcsec)
        assert_quantity_allclose(fl.locator(0.3, 0.9)[0], [0.4, 0.6, 0.8] * u.arcsec)

    @pytest.mark.parametrize(
        ("spacing", "string"),
        [
            (2 * u.deg, "15\xb0"),
            (2 * u.arcmin, "15\xb024'"),
            (2 * u.arcsec, "15\xb023'32\""),
            (0.1 * u.arcsec, "15\xb023'32.0\""),
        ],
    )
    def test_formatter_no_format(self, spacing, string):
        fl = AngleFormatterLocator()
        assert fl.formatter([15.392231] * u.degree, spacing)[0] == string

    @pytest.mark.parametrize(
        ("format_unit", "decimal", "show_decimal_unit", "spacing", "ascii", "latex"),
        [
            (u.degree, False, True, 2 * u.degree, "15\xb0", r"$15^\circ$"),
            (
                u.degree,
                False,
                True,
                2 * u.arcmin,
                "15\xb024'",
                r"$15^\circ24{}^\prime$",
            ),
            (
                u.degree,
                False,
                True,
                2 * u.arcsec,
                "15\xb023'32\"",
                r"$15^\circ23{}^\prime32{}^{\prime\prime}$",
            ),
            (
                u.degree,
                False,
                True,
                0.1 * u.arcsec,
                "15\xb023'32.0\"",
                r"$15^\circ23{}^\prime32.0{}^{\prime\prime}$",
            ),
            (u.hourangle, False, True, 15 * u.degree, "1h", r"$1^{\mathrm{h}}$"),
            (
                u.hourangle,
                False,
                True,
                15 * u.arcmin,
                "1h02m",
                r"$1^{\mathrm{h}}02^{\mathrm{m}}$",
            ),
            (
                u.hourangle,
                False,
                True,
                15 * u.arcsec,
                "1h01m34s",
                r"$1^{\mathrm{h}}01^{\mathrm{m}}34^{\mathrm{s}}$",
            ),
            (
                u.hourangle,
                False,
                True,
                1.5 * u.arcsec,
                "1h01m34.1s",
                r"$1^{\mathrm{h}}01^{\mathrm{m}}34.1^{\mathrm{s}}$",
            ),
            (u.degree, True, True, 15 * u.degree, "15\xb0", r"$15\mathrm{^\circ}$"),
            (
                u.degree,
                True,
                True,
                0.12 * u.degree,
                "15.39\xb0",
                r"$15.39\mathrm{^\circ}$",
            ),
            (
                u.degree,
                True,
                True,
                0.0036 * u.arcsec,
                "15.392231\xb0",
                r"$15.392231\mathrm{^\circ}$",
            ),
            (u.arcmin, True, True, 15 * u.degree, "924'", r"$924\mathrm{^\prime}$"),
            (
                u.arcmin,
                True,
                True,
                0.12 * u.degree,
                "923.5'",
                r"$923.5\mathrm{^\prime}$",
            ),
            (
                u.arcmin,
                True,
                True,
                0.1 * u.arcmin,
                "923.5'",
                r"$923.5\mathrm{^\prime}$",
            ),
            (
                u.arcmin,
                True,
                True,
                0.0002 * u.arcmin,
                "923.5339'",
                r"$923.5339\mathrm{^\prime}$",
            ),
            (
                u.arcsec,
                True,
                True,
                0.01 * u.arcsec,
                '55412.03"',
                r"$55412.03\mathrm{^{\prime\prime}}$",
            ),
            (
                u.arcsec,
                True,
                True,
                0.001 * u.arcsec,
                '55412.032"',
                r"$55412.032\mathrm{^{\prime\prime}}$",
            ),
            (
                u.mas,
                True,
                True,
                0.001 * u.arcsec,
                "55412032 mas",
                r"$55412032\;\mathrm{mas}$",
            ),
            (u.degree, True, False, 15 * u.degree, "15", "15"),
            (u.degree, True, False, 0.12 * u.degree, "15.39", "15.39"),
            (u.degree, True, False, 0.0036 * u.arcsec, "15.392231", "15.392231"),
            (u.arcmin, True, False, 15 * u.degree, "924", "924"),
            (u.arcmin, True, False, 0.12 * u.degree, "923.5", "923.5"),
            (u.arcmin, True, False, 0.1 * u.arcmin, "923.5", "923.5"),
            (u.arcmin, True, False, 0.0002 * u.arcmin, "923.5339", "923.5339"),
            (u.arcsec, True, False, 0.01 * u.arcsec, "55412.03", "55412.03"),
            (u.arcsec, True, False, 0.001 * u.arcsec, "55412.032", "55412.032"),
            (u.mas, True, False, 0.001 * u.arcsec, "55412032", "55412032"),
            # Make sure that specifying None defaults to
            # decimal for non-degree or non-hour angles
            (
                u.arcsec,
                None,
                True,
                0.01 * u.arcsec,
                '55412.03"',
                r"$55412.03\mathrm{^{\prime\prime}}$",
            ),
        ],
    )
    def test_formatter_no_format_with_units(
        self, format_unit, decimal, show_decimal_unit, spacing, ascii, latex
    ):
        # Check the formatter works when specifying the default units and
        # decimal behavior to use.
        fl = AngleFormatterLocator(
            unit=u.degree,
            format_unit=format_unit,
            decimal=decimal,
            show_decimal_unit=show_decimal_unit,
        )
        assert fl.formatter([15.392231] * u.degree, spacing, format="ascii")[0] == ascii
        assert fl.formatter([15.392231] * u.degree, spacing, format="latex")[0] == latex

    def test_incompatible_unit_decimal(self):
        with pytest.raises(
            UnitsError,
            match=(
                r"Units should be degrees or hours when using non-decimal .sexagesimal."
                r" mode"
            ),
        ):
            AngleFormatterLocator(unit=u.arcmin, decimal=False)


class TestScalarFormatterLocator:
    def test_no_options(self):
        fl = ScalarFormatterLocator(unit=u.m)
        assert fl.values is None
        assert fl.number == 5
        assert fl.spacing is None

    def test_too_many_options(self):
        MESSAGE = r"At most one of values/number/spacing can be specified"

        with pytest.raises(ValueError, match=MESSAGE):
            ScalarFormatterLocator(values=[1.0, 2.0] * u.m, number=5)

        with pytest.raises(ValueError, match=MESSAGE):
            ScalarFormatterLocator(values=[1.0, 2.0] * u.m, spacing=5.0 * u.m)

        with pytest.raises(ValueError, match=MESSAGE):
            ScalarFormatterLocator(number=5, spacing=5.0 * u.m)

        with pytest.raises(ValueError, match=MESSAGE):
            ScalarFormatterLocator(values=[1.0, 2.0] * u.m, number=5, spacing=5.0 * u.m)

    def test_values(self):
        fl = ScalarFormatterLocator(values=[0.1, 1.0, 14.0] * u.m, unit=u.m)
        assert fl.values.value.tolist() == [0.1, 1.0, 14.0]
        assert fl.number is None
        assert fl.spacing is None

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(values.value, [0.1, 1.0, 14.0])

    def test_number(self):
        fl = ScalarFormatterLocator(number=7, unit=u.m)
        assert fl.values is None
        assert fl.number == 7
        assert fl.spacing is None

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(values.value, np.linspace(36.0, 54.0, 10))

        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.value, np.linspace(34.4, 36, 9))

        fl.format = "x"
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.value, [35.0, 36.0])

    def test_spacing(self):
        fl = ScalarFormatterLocator(spacing=3.0 * u.m)
        assert fl.values is None
        assert fl.number is None
        assert fl.spacing == 3.0 * u.m

        values, spacing = fl.locator(34.3, 55.4)
        assert_almost_equal(values.value, [36.0, 39.0, 42.0, 45.0, 48.0, 51.0, 54.0])

        fl.spacing = 0.5 * u.m
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.value, [34.5, 35.0, 35.5, 36.0])

        with pytest.warns(UserWarning, match=r"Spacing is too small"):
            fl.format = "x"
        values, spacing = fl.locator(34.3, 36.1)
        assert_almost_equal(values.value, [35.0, 36.0])

    def test_minor_locator(self):
        fl = ScalarFormatterLocator(unit=u.m)

        values, spacing = fl.locator(34.3, 55.4)

        minor_values = fl.minor_locator(spacing, 5, 34.3, 55.4)

        assert_almost_equal(
            minor_values.value,
            [
                36.0,
                37.0,
                38.0,
                39.0,
                41.0,
                42.0,
                43.0,
                44.0,
                46.0,
                47.0,
                48.0,
                49.0,
                51.0,
                52.0,
                53.0,
                54.0,
            ],
        )

        minor_values = fl.minor_locator(spacing, 2, 34.3, 55.4)

        assert_almost_equal(minor_values.value, [37.5, 42.5, 47.5, 52.5])

        fl.values = [0.1, 1.0, 14.0] * u.m

        values, spacing = fl.locator(34.3, 36.1)

        minor_values = fl.minor_locator(spacing, 2, 34.3, 55.4)

        assert_almost_equal(minor_values.value, [])

    @pytest.mark.parametrize(
        ("format", "string"),
        [
            ("x", "15"),
            ("x.x", "15.4"),
            ("x.xx", "15.39"),
            ("x.xxx", "15.392"),
            ("%g", "15.3922"),
            ("%f", "15.392231"),
            ("%.2f", "15.39"),
            ("%.3f", "15.392"),
        ],
    )
    def test_format(self, format, string):
        fl = ScalarFormatterLocator(number=5, format=format, unit=u.m)
        assert fl.formatter([15.392231] * u.m, None)[0] == string

    @pytest.mark.parametrize(
        ("format", "string"),
        [("x", "1539"), ("x.x", "1539.2"), ("x.xx", "1539.22"), ("x.xxx", "1539.223")],
    )
    def test_format_unit(self, format, string):
        fl = ScalarFormatterLocator(number=5, format=format, unit=u.m)
        fl.format_unit = u.cm
        assert fl.formatter([15.392231] * u.m, None)[0] == string

    @pytest.mark.parametrize("format", ["dd", "dd:mm", "xx:mm", "mx.xxx"])
    def test_invalid_formats(self, format):
        fl = ScalarFormatterLocator(number=5, unit=u.m)
        with pytest.raises(ValueError, match=f"Invalid format: {format}"):
            fl.format = format

    @pytest.mark.parametrize(
        ("format", "base_spacing"),
        [("x", 1.0 * u.m), ("x.x", 0.1 * u.m), ("x.xxx", 0.001 * u.m)],
    )
    def test_base_spacing(self, format, base_spacing):
        fl = ScalarFormatterLocator(number=5, format=format, unit=u.m)
        assert fl.base_spacing == base_spacing

    def test_incorrect_spacing(self):
        fl = ScalarFormatterLocator(unit=u.m)
        fl.spacing = 0.032 * u.m
        with pytest.warns(
            UserWarning, match=r"Spacing is not a multiple of base spacing"
        ):
            fl.format = "x.xx"
        assert_almost_equal(fl.spacing.to_value(u.m), 0.03)

    def test_values_unit(self):
        # Make sure that the intrinsic unit and format unit are correctly
        # taken into account when using the locator

        fl = ScalarFormatterLocator(unit=u.cm, format_unit=u.m)
        assert_quantity_allclose(
            fl.locator(850, 2150)[0],
            [1000.0, 1200.0, 1400.0, 1600.0, 1800.0, 2000.0] * u.cm,
        )

        fl = ScalarFormatterLocator(unit=u.cm, format_unit=u.m)
        fl.format = "x.x"
        assert_quantity_allclose(fl.locator(1, 19)[0], [10] * u.cm)



"""Implements the wrapper for the Astropy test runner.

This is for backward-compatibility for other downstream packages and can be removed
once astropy-helpers has reached end-of-life.

"""
import os
import shutil
import stat
import subprocess
import sys
import tempfile
from contextlib import contextmanager

from setuptools import Command

from astropy.logger import log


@contextmanager
def _suppress_stdout():
    """
    A context manager to temporarily disable stdout.

    Used later when installing a temporary copy of astropy to avoid a
    very verbose output.
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


class FixRemoteDataOption(type):
    """
    This metaclass is used to catch cases where the user is running the tests
    with --remote-data. We've now changed the --remote-data option so that it
    takes arguments, but we still want --remote-data to work as before and to
    enable all remote tests. With this metaclass, we can modify sys.argv
    before setuptools try to parse the command-line options.
    """

    def __init__(cls, name, bases, dct):
        try:
            idx = sys.argv.index("--remote-data")
        except ValueError:
            pass
        else:
            sys.argv[idx] = "--remote-data=any"

        try:
            idx = sys.argv.index("-R")
        except ValueError:
            pass
        else:
            sys.argv[idx] = "-R=any"

        return super().__init__(name, bases, dct)


class AstropyTest(Command, metaclass=FixRemoteDataOption):
    description = "Run the tests for this package"

    user_options = [
        (
            "package=",
            "P",
            "The name of a specific package to test, e.g. 'io.fits' or 'utils'. "
            "Accepts comma separated string to specify multiple packages. "
            "If nothing is specified, all default tests are run.",
        ),
        (
            "test-path=",
            "t",
            "Specify a test location by path.  If a relative path to a  .py file, "
            'it is relative to the built package, so e.g., a  leading "astropy/" '
            "is necessary.  If a relative  path to a .rst file, it is relative to "
            "the directory *below* the --docs-path directory, so a leading "
            '"docs/" is usually necessary.  May also be an absolute path.',
        ),
        ("verbose-results", "V", "Turn on verbose output from pytest."),
        ("plugins=", "p", "Plugins to enable when running pytest."),
        ("pastebin=", "b", "Enable pytest pastebin output. Either 'all' or 'failed'."),
        ("args=", "a", "Additional arguments to be passed to pytest."),
        (
            "remote-data=",
            "R",
            "Run tests that download remote data. Should be "
            "one of none/astropy/any (defaults to none).",
        ),
        (
            "pep8",
            "8",
            "Enable PEP8 checking and disable regular tests. "
            "Requires the pytest-pep8 plugin.",
        ),
        ("pdb", "d", "Start the interactive Python debugger on errors."),
        ("coverage", "c", "Create a coverage report. Requires the coverage package."),
        (
            "parallel=",
            "j",
            "Run the tests in parallel on the specified number of "
            'CPUs.  If "auto", all the cores on the machine will be '
            "used.  Requires the pytest-xdist plugin.",
        ),
        (
            "docs-path=",
            None,
            "The path to the documentation .rst files.  If not provided, and "
            'the current directory contains a directory called "docs", that '
            "will be used.",
        ),
        ("skip-docs", None, "Don't test the documentation .rst files."),
        (
            "repeat=",
            None,
            "How many times to repeat each test (can be used to check for "
            "sporadic failures).",
        ),
        (
            "temp-root=",
            None,
            "The root directory in which to create the temporary testing files. "
            "If unspecified the system default is used (e.g. /tmp) as explained "
            "in the documentation for tempfile.mkstemp.",
        ),
        (
            "verbose-install",
            None,
            "Turn on terminal output from the installation of astropy in a "
            "temporary folder.",
        ),
        ("readonly", None, "Make the temporary installation being tested read-only."),
    ]

    package_name = ""

    def initialize_options(self):
        self.package = None
        self.test_path = None
        self.verbose_results = False
        self.plugins = None
        self.pastebin = None
        self.args = None
        self.remote_data = "none"
        self.pep8 = False
        self.pdb = False
        self.coverage = False
        self.parallel = 0
        self.docs_path = None
        self.skip_docs = False
        self.repeat = None
        self.temp_root = None
        self.verbose_install = False
        self.readonly = False

    def finalize_options(self):
        # Normally we would validate the options here, but that's handled in
        # run_tests
        pass

    def generate_testing_command(self):
        """
        Build a Python script to run the tests.
        """
        cmd_pre = ""  # Commands to run before the test function
        cmd_post = ""  # Commands to run after the test function

        if self.coverage:
            pre, post = self._generate_coverage_commands()
            cmd_pre += pre
            cmd_post += post

        set_flag = "import builtins; builtins._ASTROPY_TEST_ = True"

        cmd = (  # see _build_temp_install below
            "{cmd_pre}{0}; import {1.package_name}, sys; result = ("
            "{1.package_name}.test("
            "package={1.package!r}, "
            "test_path={1.test_path!r}, "
            "args={1.args!r}, "
            "plugins={1.plugins!r}, "
            "verbose={1.verbose_results!r}, "
            "pastebin={1.pastebin!r}, "
            "remote_data={1.remote_data!r}, "
            "pep8={1.pep8!r}, "
            "pdb={1.pdb!r}, "
            "parallel={1.parallel!r}, "
            "docs_path={1.docs_path!r}, "
            "skip_docs={1.skip_docs!r}, "
            "add_local_eggs_to_path=True, "
            "repeat={1.repeat!r})); "
            "{cmd_post}"
            "sys.exit(result)"
        )
        return cmd.format(set_flag, self, cmd_pre=cmd_pre, cmd_post=cmd_post)

    def run(self):
        """Run the tests!"""
        # Install the runtime dependencies.
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(self.distribution.install_requires)

        # Ensure there is a doc path
        if self.docs_path is None:
            cfg_docs_dir = self.distribution.get_option_dict("build_docs").get(
                "source_dir", None
            )

            # Some affiliated packages use this.
            # See astropy/package-template#157
            if cfg_docs_dir is not None and os.path.exists(cfg_docs_dir[1]):
                self.docs_path = os.path.abspath(cfg_docs_dir[1])

            # fall back on a default path of "docs"
            elif os.path.exists("docs"):  # pragma: no cover
                self.docs_path = os.path.abspath("docs")

        # Build a testing install of the package
        self._build_temp_install()

        # Install the test dependencies
        # NOTE: we do this here after _build_temp_install because there is
        # a weird but which occurs if psutil is installed in this way before
        # astropy is built, Cython can have segmentation fault. Strange, eh?
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)

        # Copy any additional dependencies that may have been installed via
        # tests_requires or install_requires. We then pass the
        # add_local_eggs_to_path=True option to package.test() to make sure the
        # eggs get included in the path.
        if os.path.exists(".eggs"):
            shutil.copytree(".eggs", os.path.join(self.testing_path, ".eggs"))

        # This option exists so that we can make sure that the tests don't
        # write to an installed location.
        if self.readonly:
            log.info("changing permissions of temporary installation to read-only")
            self._change_permissions_testing_path(writable=False)

        # Run everything in a try: finally: so that the tmp dir gets deleted.
        try:
            # Construct this modules testing command
            cmd = self.generate_testing_command()

            # Run the tests in a subprocess--this is necessary since
            # new extension modules may have appeared, and this is the
            # easiest way to set up a new environment

            testproc = subprocess.Popen(
                [sys.executable, "-c", cmd], cwd=self.testing_path, close_fds=False
            )
            retcode = testproc.wait()
        except KeyboardInterrupt:
            import signal

            # If a keyboard interrupt is handled, pass it to the test
            # subprocess to prompt pytest to initiate its teardown
            testproc.send_signal(signal.SIGINT)
            retcode = testproc.wait()
        finally:
            # Remove temporary directory
            if self.readonly:
                self._change_permissions_testing_path(writable=True)
            shutil.rmtree(self.tmp_dir)

        raise SystemExit(retcode)

    def _build_temp_install(self):
        """
        Install the package and to a temporary directory for the purposes of
        testing. This allows us to test the install command, include the
        entry points, and also avoids creating pyc and __pycache__ directories
        inside the build directory.
        """
        # On OSX the default path for temp files is under /var, but in most
        # cases on OSX /var is actually a symlink to /private/var; ensure we
        # dereference that link, because pytest is very sensitive to relative
        # paths...

        tmp_dir = tempfile.mkdtemp(
            prefix=self.package_name + "-test-", dir=self.temp_root
        )
        self.tmp_dir = os.path.realpath(tmp_dir)

        log.info(f"installing to temporary directory: {self.tmp_dir}")

        # We now install the package to the temporary directory. We do this
        # rather than build and copy because this will ensure that e.g. entry
        # points work.
        self.reinitialize_command("install")
        install_cmd = self.distribution.get_command_obj("install")
        install_cmd.prefix = self.tmp_dir
        if self.verbose_install:
            self.run_command("install")
        else:
            with _suppress_stdout():
                self.run_command("install")

        # We now get the path to the site-packages directory that was created
        # inside self.tmp_dir
        install_cmd = self.get_finalized_command("install")
        self.testing_path = install_cmd.install_lib

        # Ideally, docs_path is set properly in run(), but if it is still
        # not set here, do not pretend it is, otherwise bad things happen.
        # See astropy/package-template#157
        if self.docs_path is not None:
            new_docs_path = os.path.join(
                self.testing_path, os.path.basename(self.docs_path)
            )
            shutil.copytree(self.docs_path, new_docs_path)
            self.docs_path = new_docs_path

        shutil.copy("setup.cfg", self.testing_path)

    def _change_permissions_testing_path(self, writable=False):
        if writable:
            basic_flags = stat.S_IRUSR | stat.S_IWUSR
        else:
            basic_flags = stat.S_IRUSR
        for root, dirs, files in os.walk(self.testing_path):
            for dirname in dirs:
                os.chmod(os.path.join(root, dirname), basic_flags | stat.S_IXUSR)
            for filename in files:
                os.chmod(os.path.join(root, filename), basic_flags)

    def _generate_coverage_commands(self):
        """
        This method creates the post and pre commands if coverage is to be
        generated.
        """
        if self.parallel != 0:
            raise ValueError("--coverage can not be used with --parallel")

        try:
            import coverage  # noqa: F401
        except ImportError:
            raise ImportError(
                "--coverage requires that the coverage package is installed."
            )

        # Don't use get_pkg_data_filename here, because it
        # requires importing astropy.config and thus screwing
        # up coverage results for those packages.
        coveragerc = os.path.join(
            self.testing_path,
            self.package_name.replace(".", "/"),
            "tests",
            "coveragerc",
        )

        with open(coveragerc) as fd:
            coveragerc_content = fd.read()

        coveragerc_content = coveragerc_content.replace(
            "{packagename}", self.package_name.replace(".", "/")
        )
        tmp_coveragerc = os.path.join(self.tmp_dir, "coveragerc")
        with open(tmp_coveragerc, "wb") as tmp:
            tmp.write(coveragerc_content.encode("utf-8"))

        cmd_pre = (
            "import coverage; cov ="
            f' coverage.coverage(data_file=r"{os.path.abspath(".coverage")}",'
            f' config_file=r"{os.path.abspath(tmp_coveragerc)}"); cov.start();'
        )
        cmd_post = (
            "cov.stop(); from astropy.tests.helper import _save_coverage;"
            f' _save_coverage(cov, result, r"{os.path.abspath(".")}",'
            f' r"{os.path.abspath(self.testing_path)}");'
        )

        return cmd_pre, cmd_post



# Licensed under a 3-clause BSD style license

import os

import numpy
from setuptools import Extension

ROOT = os.path.relpath(os.path.dirname(__file__))


def get_extensions():
    sources = [
        os.path.join(ROOT, "cparser.pyx"),
        os.path.join(ROOT, "src", "tokenizer.c"),
    ]
    ascii_ext = Extension(
        name="astropy.io.ascii.cparser",
        include_dirs=[numpy.get_include()],
        sources=sources,
    )
    return [ascii_ext]
